#!/usr/bin/python3

# Copyright (C) 2020  ASTRON (Netherlands Institute for Radio Astronomy)
# P.O. Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This file is part of the LOFAR software suite.
# The LOFAR software suite is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# The LOFAR software suite is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with the LOFAR software suite. If not, see <http://www.gnu.org/licenses/>.

from dateutil import parser
from lofar.sas.tmss.tmss.tmssapp.models import *
from lofar.sas.tmss.tmss.tmssapp.conversions import antennafields_for_antennaset_and_station
from lofar.parameterset import parameterset
from lofar.common.util import single_line_with_single_spaces
from lofar.sas.tmss.tmss.exceptions import SubtaskInvalidStateException, SubtaskException

import logging
logger = logging.getLogger(__name__)

def process_feedback_into_subtask_dataproducts(subtask:Subtask, feedback: parameterset) -> Subtask:
    if subtask.specifications_template.type.value not in [SubtaskType.Choices.OBSERVATION.value, SubtaskType.Choices.PIPELINE.value]:
            raise SubtaskException("Cannot process feedback for subtask id=%s since type=%s not in %s" %
                                   (subtask.id, subtask.specifications_template.type.value,
                                    [SubtaskType.Choices.OBSERVATION.value, SubtaskType.Choices.PIPELINE.value]))

    if subtask.state.value != SubtaskState.objects.get(value='finishing').value:
        raise SubtaskInvalidStateException("Cannot process feedback for subtask id=%s because the state is '%s' and not '%s'" % (subtask.id, subtask.state.value, SubtaskState.Choices.FINISHING.value))

    logger.info('processing feedback into the dataproducts of subtask id=%s type=%s feedback:\n%s', subtask.id, subtask.specifications_template.type.value, str(feedback))

    # create a subset in dict-form with the dataproduct information
    if subtask.specifications_template.type.value == SubtaskType.Choices.OBSERVATION.value:
        dataproducts_feedback = feedback.makeSubset('Observation.DataProducts.')
    elif subtask.specifications_template.type.value == SubtaskType.Choices.PIPELINE.value:
        dataproducts_feedback = feedback.makeSubset('LOFAR.ObsSW.Observation.DataProducts.')

    # extract the unique dataproduct keys, so we can loop over them
    dp_keys = sorted(list(set([key[:key.find('.')] for key in dataproducts_feedback.keys() if key.startswith('Output_')])))

    # process each dataproduct subset
    for dp_key in dp_keys:
        dp_feedback = dataproducts_feedback.makeSubset(dp_key+'.').dict()

        # determine corresponding TMSS dataproduct
        try:
            dataproduct = subtask.output_dataproducts.get(filename=dp_feedback.get('filename'))
        except Exception as e:
            logger.error("cannot process feedback: %s. No output dataproduct known for subtask id=%s feedback: %s", e, subtask.id, dp_feedback)
            continue

        try:
            logger.info('processing feedback for dataproduct id=%s filename=%s of subtask id=%s feedback: %s', dataproduct.id, dataproduct.filename, subtask.id, single_line_with_single_spaces(str(dp_feedback)))

            # set the feedback_template, so we can fill the feedback json doc for the dataproduct
            dataproduct.feedback_template = DataproductFeedbackTemplate.objects.get(name='feedback')

            # derive values or collect for different subtask types
            storagewriter = dp_feedback['storageWriter'].lower()
            if storagewriter == "casa":
                storagewriter = "standard"    # todo: is that correct?
            elif storagewriter == "lofar":
                storagewriter = "lofarstman"

            if subtask.specifications_template.type.value == SubtaskType.Choices.OBSERVATION.value:
                subbands = [int(dp_feedback['stationSubband'])]
                duration = (subtask.stop_time - subtask.start_time).total_seconds()
                antennaset = subtask.specifications_doc['stations']['antenna_set']
                stationlist = subtask.specifications_doc['stations']['station_list']
                antennatype = antennaset.split('_')[0]  # LBA or HBA
                antennafields = []
                for station in stationlist:
                    fields = antennafields_for_antennaset_and_station(antennaset, station)
                    antennafields += [{"station": station, "field": field, "type": antennatype} for field in fields]
                pointing = subtask.specifications_doc['stations']['digital_pointings'][int(dp_feedback['SAP'])]['pointing']
            elif subtask.specifications_template.type.value == SubtaskType.Choices.PIPELINE.value:
                input_dataproduct = DataproductTransform.objects.get(output=dataproduct).input
                logger.debug('Found input dataproduct %s' % input_dataproduct.filename)
                subbands = input_dataproduct.feedback_doc["frequency"]['subbands']
                duration = float(dp_feedback['duration'])
                antennaset = input_dataproduct.feedback_doc["antennas"]['set']
                antennafields = input_dataproduct.feedback_doc["antennas"]['fields']
                pointing = input_dataproduct.feedback_doc["target"]['pointing']


            # add feedback doc to dataproduct
            dataproduct.feedback_doc={
                "percentage_written": int(dp_feedback['percentageWritten']),
                "frequency": {
                    "subbands": subbands,
                    "central_frequencies": [float(dp_feedback['centralFrequency'])],
                    "channel_width": float(dp_feedback['channelWidth']),
                    "channels_per_subband": int(dp_feedback['channelsPerSubband'])
                },
                "time": {
                    "start_time": parser.parse(dp_feedback['startTime'], ignoretz=True).isoformat()+'Z',
                    "duration": duration,
                    "sample_width": float(dp_feedback['integrationInterval']),
                },
                "antennas": {
                    "set": antennaset,
                    "fields": antennafields
                },
                "target": {
                    "pointing": pointing
                },
                "samples": {
                    "polarisations": ["XX","XY","YX","YY"],         # fixed
                    "type": "float",                                # fixed
                    "bits": 32,                                     # fixed
                    "writer": storagewriter,
                    "writer_version": dp_feedback['storageWriterVersion'],
                    "complex": True                                 # fixed
                }
            }

            dataproduct.save()
            logger.info('saved processed feedback into dataproduct id=%s filename=%s feedback_doc=%s', dataproduct.id, dataproduct.filename, dataproduct.feedback_doc)
        except Exception as e:
            logger.error('error while processing feedback for dataproduct id=%s filename=%s feedback=%s error: %s', dataproduct.id, dataproduct.filename, dp_feedback, e)

    subtask.refresh_from_db()
    return subtask


def append_to_subtask_raw_feedback(subtask: Subtask, feedback: parameterset) -> Subtask:
    """ append/merge the given feedback into the already stored raw_feedback
    by using this parset-merging strategy we ensure a properly formatted plain text raw_feedback doc with no duplicate entries
    """
    # check we are in finishing state
    if subtask.state.value != SubtaskState.objects.get(value='finishing').value:
        raise SubtaskInvalidStateException("Cannot process feedback for subtask id=%s because the state is '%s' and not '%s'" % (subtask.id, subtask.state.value, SubtaskState.Choices.FINISHING.value))

    feedback_parset = parameterset.fromString(subtask.raw_feedback or "")
    feedback_parset.adoptDict(feedback.dict())
    subtask.raw_feedback = str(feedback_parset)
    subtask.save()
    return subtask

def process_feedback_for_subtask_and_set_to_finished_if_complete(subtask: Subtask, feedback_doc: str) -> Subtask:
    """
    Translates raw feedback from a subtask (which has been provided by Cobalt or pipelines) and translate it to
    json documents for the individual dataproducts.
    """
    if subtask.state.value == SubtaskState.objects.get(value='started').value:
        logger.info("received feedback for subtask id=%s while it is still running (state=%s). Setting state to 'finishing' and stop_time to now.", subtask.id, subtask.state.value)
        subtask.state = SubtaskState.objects.get(value='finishing')
        subtask.stop_time = datetime.utcnow()
        subtask.save()

    # the submitted feedback_doc is (should be) a plain text document in parset format
    # so, treat it as a parset
    new_feedback_parset = parameterset.fromString(feedback_doc)

    # store the new feedback
    subtask = append_to_subtask_raw_feedback(subtask=subtask, feedback=new_feedback_parset)

    # and process it
    subtask = process_feedback_into_subtask_dataproducts(subtask, new_feedback_parset)

    # if complete, set subtask state to finished
    if subtask.is_feedback_complete:
        logger.info("Feedback for subtask id=%s is complete. Setting state to 'finished'", subtask.id)
        subtask.state = SubtaskState.objects.get(value='finished')
        subtask.save()

    return subtask


def reprocess_raw_feedback_for_subtask_and_set_to_finished_if_complete(subtask: Subtask) -> Subtask:
    """
    Reprocesses the stored raw feedback from a subtask (which has been provided by Cobalt or pipelines) and translate it to
    json documents for the individual dataproducts.
    """
    if not subtask.raw_feedback:
        raise SubtaskException("Cannot reprocess raw_feedback for subtask id=%s because it is empty" % (subtask.id))

    try:
        raw_feedback_parset = parameterset.fromString(subtask.raw_feedback)
    except Exception as e:
        raise SubtaskException("Cannot reprocess raw_feedback for subtask id=%s because it cannot be interpreted as parset. Error=%s\n%s" % (subtask.id, str(e), subtask.raw_feedback))

    if subtask.state.value in (SubtaskState.objects.get(value='started').value, SubtaskState.objects.get(value='finished').value):
        # allow started/running subtask to go to finishing, (re)process, and then go to finished if complete
        # allow finished subtask to go back to finishing, reprocess, and then go to finished again if complete
        subtask.state = SubtaskState.objects.get(value='finishing')
        subtask.save()

    subtask = process_feedback_into_subtask_dataproducts(subtask, raw_feedback_parset)

    # if complete, set subtask state to finished
    if subtask.is_feedback_complete:
        logger.info("Feedback for subtask id=%s is complete. Setting state to 'finished'", subtask.id)
        subtask.state = SubtaskState.objects.get(value='finished')

    subtask.save()
    return subtask


