from django.db.models import Sum
from lofar.sas.tmss.tmss.tmssapp import models
from lofar.sas.tmss.tmss.tmssapp import serializers

from rest_framework.request import Request
from datetime import timedelta


def create_project_report(request: Request, project: models.Project) -> {}:
    """
    Create a project report as a JSON object.
    """
    result = {'project': project.pk}
    result['quota'] = _get_quotas_from_project(request, project.pk)
    result['SUBs'], result['durations'] = _get_subs_and_durations_from_project(project.pk)
    result['LTA dataproducts'] = _get_lta_dataproducts(project.name)
    result['SAPs'] = _get_saps(project.pk)

    return result


def _get_quotas_from_project(request: Request, project_pk: int) -> []:
    """
    Help function to retrieve quotas.
    """
    project_quotas = models.ProjectQuota.objects.filter(project=project_pk)
    project_quotas_data = [serializers.ProjectQuotaSerializer(pq, context={'request': request}).data for pq in project_quotas]
    quotas = [{k: pqd[k] for k in ('id', 'resource_type_id', 'value')} for pqd in project_quotas_data]
    return quotas


def _get_subs_and_durations_from_project(project_pk: int) -> ({}, {}):
    """
    Help function to retrieve durations and scheduling_units distinguished by success/fail.
    """
    # Get SUBs related to the project
    scheduling_unit_blueprints = models.SchedulingUnitBlueprint.objects.filter(draft__scheduling_set__project__pk=project_pk)
    # TODO: Split into total, prio A, prio B? See TMSS-592.
    total_duration, total_succeeded_duration, total_failed_duration = timedelta(), timedelta(), timedelta()
    subs_succeeded, subs_failed = [], []

    # NOTE: This might be optimised later with the use of Django's ORM as done for LTA dataproducts.
    for sub in scheduling_unit_blueprints:  # Distinguish between succeeded and failed observations
        # TODO: Use QA workflow flag instead of the finished status? See TMSS-592.
        if sub.status == 'finished':        # Succeeded observations
            total_succeeded_duration += sub.duration
            subs_succeeded.append({'id': sub.pk, 'name': sub.name, 'duration': sub.duration.total_seconds()})
        elif sub.status == 'cancelled':     # Failed observations
            total_failed_duration += sub.duration
            subs_failed.append({'id': sub.pk, 'name': sub.name, 'duration': sub.duration.total_seconds()})
        total_duration += sub.duration      # Total duration without considering the status of the obs.

    total_not_cancelled = total_duration - total_failed_duration  # Calculate not_cancelled duration
    durations = {'total': total_duration.total_seconds(), 'total_succeeded': total_succeeded_duration.total_seconds(),
                 'total_not_cancelled': total_not_cancelled.total_seconds(), 'total_failed': total_failed_duration.total_seconds()}
    subs = {'finished': subs_succeeded, 'failed': subs_failed}

    return subs, durations


def _get_lta_dataproducts(project_name: str) -> {}:
    """
    Help function to retrieve the sum of the LTA dataproducts sizes.
    """
    # Query dataproducts from Subtasks of type 'ingest' within 'finished' status
    return models.Dataproduct.objects.filter(producer__subtask__specifications_template__type='ingest') \
        .filter(producer__subtask__state__value='finished') \
        .filter(producer__subtask__task_blueprints__draft__scheduling_unit_draft__scheduling_set__project__name=project_name) \
        .aggregate(Sum('size'))


def _get_saps(project_pk: int) -> []:
    """
    Help function to retrieve SAPs.
    """
    # TODO: For each unique target (SAP name) get the sum of target observation durations from the tasks.
    return [{'sap_name': 'placeholder', 'total_exposure': 0}, ]
