from django.utils.decorators import method_decorator
from django.db.models.signals import post_save
from viewflow import flow
from viewflow.models import Task
from viewflow.base import this, Flow
from viewflow.flow.views import CreateProcessView, UpdateProcessView, AssignTaskView
from viewflow.activation import FuncActivation, ViewActivation
from viewflow.flow.nodes import Signal
from viewflow import mixins

from .. import models
from .. import viewsets
from .. import forms

from lofar.sas.tmss.tmss.tmssapp.models import Subtask, SchedulingUnitBlueprint
from lofar.common.datetimeutils import round_to_second_precision

from django.dispatch import receiver
from lofar.sas.tmss.tmss.workflowapp.signals import scheduling_unit_blueprint_status_changed_signal, scheduling_unit_blueprint_cannot_proceed_signal, ingest_task_blueprint_status_changed_signal, obs_task_status_changed_signal

from viewflow import frontend, ThisObject
from viewflow.activation import STATUS
from viewflow.models import Process

from lofar.sas.tmss.client.tmssbuslistener import *

from datetime import datetime

import logging
logger = logging.getLogger(__name__)

class ConditionActivation(FuncActivation):
    @classmethod
    def activate(cls, flow_task, prev_activation, token):
        activation = super(ConditionActivation, cls).activate(flow_task, prev_activation, token)

        if flow_task.condition_check(activation, None):
            # condition holds on activation
            activation.prepare()
            activation.done()

        return activation

class Condition(Signal):
    task_type = "HUMAN" # makes it show up in the unassigned task lists
    activation_class = ConditionActivation

    def __init__(self, condition_check, signal, sender=None, task_loader=None, **kwargs):
        """
        Instantiate a Signal task.

        :param signal: A django signal to connect
        :param receiver: Callable[activation, **kwargs]
        :param sender: Optional signal sender
        :param task_loader: Callable[**kwargs] -> Task
        :param allow_skip: If True task_loader can return None if
                           signal could be skipped.

        You can skip a `task_loader` if the signal going to be
        sent with Task instance.
        """
        self.condition_check = condition_check
        super(Condition, self).__init__(signal, self.signal_handler, sender, task_loader, **kwargs)

    @method_decorator(flow.flow_signal)
    def signal_handler(self, activation, sender, instance, **signal_kwargs):

      if activation.get_status() == STATUS.DONE:
          # race condition -- condition was true on activation but we also receive the signal now
          return

      activation.prepare()
      if activation.flow_task.condition_check(activation, instance):
          activation.done()
          

    def ready(self):
        """Resolve internal `this`-references. and subscribe to the signal."""

        if isinstance(self.condition_check, ThisObject):
            self.condition_check = getattr(self.flow_class.instance, self.condition_check.name)

        super(Condition, self).ready()


@frontend.register
class SchedulingUnitFlow(Flow):
   
    process_class = models.SchedulingUnitProcess

    #STEP 1
    start = (
        flow.StartSignal(
          scheduling_unit_blueprint_status_changed_signal,
          this.on_save_can_start,
        ).Next(this.wait_scheduled)
    )

    #STEP 2
    wait_scheduled = (
        Condition(
          this.check_condition_scheduled,
          obs_task_status_changed_signal,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.wait_processed)
    )
    
    wait_processed = (
        flow.Signal(
          scheduling_unit_blueprint_cannot_proceed_signal,
          this.check_condition_processed,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.qa_reporting_to)
    )

    #STEP 1
    #QA Reporting (TO)
    qa_reporting_to = (
        flow.View(
            viewsets.QAReportingTOView,
            task_description='QA Reporting (TO)'
        ).Permission(
            auto_create=True
        ).Next(this.check_operator_accept)
    )
    
    #Quality Acceptable
    check_operator_accept = (
        flow.If(lambda activation: activation.process.qa_reporting_to.operator_accept)
        .Then(this.qa_reporting_sos)
        .Else(this.mark_sub)
    )

    #QA Reporting (SOS)
    qa_reporting_sos = (
        flow.View(
            viewsets.QAReportingSOSView,
            task_description='QA Reporting (SOS)'
        ).Permission(
            auto_create=True
        ).Next(this.check_sos_accept_show_pi)
    )

    #Quality Acceptable
    check_sos_accept_show_pi = (
        flow.If(lambda activation: activation.process.qa_reporting_sos.sos_accept_show_pi)
        .Then(this.pi_verification)
        .Else(this.mark_sub)
    )

    #PI Verification
    pi_verification = (
        flow.View(
            viewsets.PIVerificationView,
            task_description='PI Verification'
        ).Permission(
            auto_create=True
        ).Next(this.decide_acceptance)
    )

    #Decide Acceptance
    decide_acceptance = (
        flow.View(
            viewsets.DecideAcceptanceView,
            task_description='Decide Acceptance'
        ).Permission(
            auto_create=True
        ).Next(this.check_sos_accept_after_pi)
    )

    #STEP 3
    #Quality Acceptable
    check_sos_accept_after_pi = (
        flow.If(lambda activation: activation.process.decide_acceptance.sos_accept_after_pi)
        .Then(this.allow_ingest)
        .Else(this.mark_sub)
    )

    allow_ingest = (
        flow.Handler(
            this.signal_SUB_allow_ingest,
        )
        .Next(this.ingest_done)
    )

    ingest_done = (
        Condition(
          this.check_ingest_done,
          ingest_task_blueprint_status_changed_signal,
          task_loader=this.get_scheduling_unit_task
        )
        .Next(this.mark_sub)
    )

    #Mark SUB Successful/failed
    mark_sub = (
        flow.Handler(
            this.do_mark_sub
        ).Next(this.check_data_pinned)
    )

    check_data_pinned = (
        flow.If(lambda activation: activation.process.su.output_pinned)
        .Then(this.unpin_data)
        .Else(this.delete_data)
    )

    unpin_data = (
        flow.View(
            viewsets.UnpinDataView,
            task_description='Unpin Data'
        ).Permission(
            auto_create=True
        ).Next(this.delete_data)
    )

    delete_data = (
        flow.Handler(
            this.do_delete_data
        ).Next(this.end)
    )
    
    end = flow.End()
    
    @method_decorator(flow.flow_start_signal)
    def on_save_can_start(self, activation, sender, instance, status, **signal_kwargs):
        if status == "schedulable":
            try:
                process = models.SchedulingUnitProcess.objects.get(su=instance)
            except Process.DoesNotExist:
                activation.prepare()
                activation.process.su = instance
                activation.done()
                logger.info("workflow started for scheduling unit id=%s name='%s'", instance.id, instance.name)
                
            except Process.MultipleObjectsReturned:
                logger.info("QA Workflow for process %s already exists",process)

        return activation

    def do_mark_sub(self, activation):

        activation.process.su.output_pinned = True
        activation.process.su.results_accepted = ((activation.process.qa_reporting_to is not None and activation.process.qa_reporting_to.operator_accept) 
            and (activation.process.qa_reporting_sos is not None and activation.process.qa_reporting_sos.sos_accept_show_pi)
            and (activation.process.decide_acceptance is not None and activation.process.decide_acceptance.sos_accept_after_pi))
        
        activation.process.su.save()
        activation.process.save()
        return activation

    def do_delete_data(self, activation):
        activation.process.su.save()
        activation.process.save()
        return activation

    def check_condition_scheduled(self, activation, instance):
        if instance is None:
            instance = activation.process.su
        
        condition = instance.status == "scheduled"
        return condition

    @method_decorator(flow.flow_signal)
    def check_condition_processed(self, activation, **signal_kwargs):
        activation.prepare()
        activation.done()
    
    def signal_SUB_allow_ingest(self, activation):       

        logger.info("granting ingest permission for scheduling unit blueprint id=%s", activation.process.su.id)
        activation.process.su.ingest_permission_granted_since = round_to_second_precision(datetime.utcnow())
        activation.process.su.ingest_permission_required = True
        activation.process.su.save()

        activation.process.save()
    
    
    def check_ingest_done(self, activation, instance):
        if instance is None:
            instance = activation.process.su
        
        logger.info("[check ingest done] checking on %s, status %s", instance, instance.status)

        condition = instance.status == "finished"
        return condition


    def get_scheduling_unit_task(self, flow_task, sender, instance, **kwargs):
        process = models.SchedulingUnitProcess.objects.get(su=instance)
        return Task.objects.get(process=process,flow_task=flow_task)
